using System;
using System.Linq;
using System.Xml;

namespace Rsdn.SmartApp.Configuration
{
	using SectionRegSvc = IRegKeyedElementsService<Type, ConfigSectionInfo>;
	using System.Collections.Generic;
	using System.Xml.Linq;

	/// <summary>
	///   <see cref="IConfigService"/>
	/// </summary>
	public class ConfigService : IConfigService, IDisposable
	{
		private readonly IConfigDataProvider _dataProvider;
		private readonly IServiceProvider _serviceProvider;
		private readonly SectionRegSvc _regSvc;
		private readonly ConfigSerializer _serializer;
		private readonly ElementsCache<ConfigSectionInfo, object> _sections;

		/// <summary>
		///      .
		/// </summary>
		public static void CreateService(IServicePublisher publisher,
			IExtensionManager extensionManager, Type[] types, string fileName,
			IDictionary<string, string> externalVars)
		{
			CreateService(publisher, extensionManager, types,
				new FileConfigDataProvider(fileName), externalVars);
		}

		/// <summary>
		///      <see cref="IConfigDataProvider"/>.
		/// </summary>
		public static void CreateService(IServicePublisher publisher,
			IExtensionManager extensionManager, Type[] types,
			IConfigDataProvider dataProvider,
			IDictionary<string, string> externalVars)
		{
			extensionManager.Scan(types, new ConfigSectionStrategy(publisher));
			ConfigService cfgSvc = new ConfigService(publisher, dataProvider, externalVars);
			publisher.Publish<IConfigService>(cfgSvc);
		}

		/// <summary>
		///     .
		/// </summary>
		public ConfigService(IServiceProvider serviceProvider,
			IConfigDataProvider dataProvider, IDictionary<string, string> externalVars)
		{
			_serviceProvider = serviceProvider;
			_dataProvider = dataProvider;
			_serializer = new ConfigSerializer(_dataProvider, externalVars);
			_regSvc = _serviceProvider.GetService<SectionRegSvc>();
			if (_regSvc == null)
				throw new Exception("Service Rsdn.SmartApp.Configuration."
					+ "IRegKeyedElementsService<string, ConfigSectionInfo> is not published");
			_sections = new ElementsCache<ConfigSectionInfo, object>(DeserializeSection);
			_serializer.ConfigChanged += SerializerConfigChanged;
		}

		/// <summary>
		///      .
		/// </summary>
		public ConfigService(IServiceProvider serviceProvider, string configFile,
			IDictionary<string, string> externalVars)
			: this(serviceProvider, new FileConfigDataProvider(configFile), externalVars)
		{}

		/// <summary>
		///    .
		/// </summary>
		public event EventHandler<IConfigService> ConfigChanged;

		/// <summary>
		///    <see cref="ConfigChanged"/>
		/// </summary>
		protected virtual void OnConfigChanged()
		{
			if (ConfigChanged != null)
				ConfigChanged(this);
		}

		/// <summary>
		///   .
		/// </summary>
		public T GetSection<T>()
		{
			var contract = typeof(T);
			if (!_regSvc.ContainsElement(contract))
				throw new ArgumentException("Section of type '" + contract.FullName + "' not exists");
			ConfigSectionInfo secInfo = _regSvc.GetElement(contract);
			return (T)_sections.Get(secInfo);
		}

		private object DeserializeSection(ConfigSectionInfo sectionInfo)
		{
			var elems = _serializer.GetSectionContent(XName.Get(sectionInfo.Name, sectionInfo.Namespace));
			if (elems.Length > 0)
			{
				var elem = elems[0];
				if (elems.Length > 1)
					if (!sectionInfo.AllowMerge)
						throw new ApplicationException(string.Format("Section '{0}' does not allow merge", sectionInfo.Name));
					else
					{
						for (int i = 1; i < elems.Length; i++)
							elem.Add(elems[i].Nodes().ToArray());
					}
				return sectionInfo.Serializer.Deserialize(elem);
			}
			else
				return sectionInfo.Serializer.CreateDefaultSection();
		}

		private void SerializerConfigChanged(ConfigSerializer sender)
		{
			_sections.Reset();
			OnConfigChanged();
		}

		#region IDisposable Members
		/// <summary>
		/// See <see cref="IDisposable.Dispose"/>
		/// </summary>
		public void Dispose()
		{
			_serializer.Dispose();
		}
		#endregion
	}
}
