using System;
using System.Collections.Generic;

namespace Rsdn.SmartApp
{
	/// <summary>
	///   <see cref="IServicePublisher"/>
	/// </summary>
	public class ServiceManager : IServicePublisher, IServiceProvider
	{
		private readonly IServiceProvider[] _parentProviders;
		private readonly Dictionary<Type, IServiceHandler> _byTypeMap =
			new Dictionary<Type, IServiceHandler>();
		private readonly Dictionary<ServiceCookie, Type> _byCookieMap =
			new Dictionary<ServiceCookie, Type>();
		private readonly object _changeLockFlag = new object();

		/// <summary>
		///     <see cref="IServicePublisher"/>.
		/// </summary>
		/// <param name="parentProviders"> - </param>
		public ServiceManager(params IServiceProvider[] parentProviders)
			: this(false, parentProviders)
		{}

		/// <summary>
		///  .
		/// </summary>
		/// <param name="parentProviders"> - </param>
		/// <param name="publishPublisher">  
		/// <see cref="IServicePublisher"/></param>
		public ServiceManager(bool publishPublisher, params IServiceProvider[] parentProviders)
		{
			_parentProviders = parentProviders;
			int i = 0;
			foreach (IServiceProvider provider in parentProviders)
			{
				if (provider == null)
					throw new ArgumentNullException("parentProviders[" + i + "]");
				i++;
			}
			if (publishPublisher)
				Publish<IServicePublisher>(this);
		}

		/// <summary>
		///     .
		/// </summary>
		public IServiceCookie Publish<T>(T serviceInstance) where T : class
		{
			if (serviceInstance == null)
				throw new ArgumentNullException("serviceInstance");
			return Publish<T>(new SvcInstanceHandler<T>(serviceInstance));
		}

		/// <summary>
		///       .
		/// </summary>
		public IServiceCookie Publish<T>(ServiceCreator<T> serviceCreator) where T : class
		{
			if (serviceCreator == null)
				throw new ArgumentNullException("servicereator");
			return Publish<T>(new SvcCreatorHandler<T>(serviceCreator));
		}

		private IServiceCookie Publish<T>(IServiceHandler handler)
		{
			return PublishUntyped(typeof (T), handler);
		}
		
		private IServiceCookie PublishUntyped(Type svcType, IServiceHandler handler)
		{
			if (_byTypeMap.ContainsKey(svcType))
				throw new ArgumentException("Service " + svcType.FullName
					+ " already published");
			lock (_changeLockFlag)
			{
				ServiceCookie cookie = new ServiceCookie();
				_byTypeMap.Add(svcType, handler);
				_byCookieMap.Add(cookie, svcType);
				return cookie;
			}
		}

		/// <summary>
		///   .
		/// </summary>
		public void Unpublish(IServiceCookie cookie)
		{
			ServiceCookie svcCookie = cookie as ServiceCookie;
			if (!_byCookieMap.ContainsKey(svcCookie))
				throw new ArgumentException("Supplied cookie is invalid", "cookie");
			lock (_changeLockFlag)
			{
				_byTypeMap.Remove(_byCookieMap[svcCookie]);
				_byCookieMap.Remove(svcCookie);
			}
		}

		/// <summary>
		///  ,   T
		/// </summary>
		public T GetService<T>() where T : class
		{
			IServiceHandler handler;
			if (!_byTypeMap.TryGetValue(typeof (T), out handler))
				return GetParentService<T>();
			return (T)handler.GetInstance();
		}

		/// <summary>
		///     .
		/// </summary>
		public IEnumerable<object> GetCreatedInstances()
		{
			foreach (var handler in _byTypeMap.Values)
				if (handler.IsInstanceCreated)
					yield return handler.GetInstance();
		}

		private T GetParentService<T>() where T : class
		{
			if (_parentProviders != null)
				foreach (IServiceProvider sp in _parentProviders)
				{
					T svc = sp.GetService<T>();
					if (svc != null)
						return svc;
				}
			return null;
		}

		#region ServiceCookie class
		private class ServiceCookie : IServiceCookie
		{
			private readonly Guid _guid = Guid.NewGuid();

			public override int GetHashCode()
			{
				return _guid.GetHashCode();
			}

			public override bool Equals(object obj)
			{
				if (this == obj)
					return true;
				ServiceCookie serviceCookie = obj as ServiceCookie;
				if (serviceCookie == null)
					return false;
				return _guid == serviceCookie._guid;
			}
		}
		#endregion

		#region Service handlers
		private interface IServiceHandler
		{
			bool IsInstanceCreated { get; }
			object GetInstance();
		}
		
		private class SvcInstanceHandler<T> : IServiceHandler
		{
			private readonly T _svcInstance;

			public SvcInstanceHandler(T svcInstance)
			{
				_svcInstance = svcInstance;
			}

			public bool IsInstanceCreated
			{
				get { return true; }
			}

			public object GetInstance()
			{
				return _svcInstance;
			}
		}

		private class SvcCreatorHandler<T> : IServiceHandler
		{
			private readonly ServiceCreator<T> _serviceCreator;
			private T _svcInstance;

			public SvcCreatorHandler(ServiceCreator<T> serviceCreator)
			{
				_serviceCreator = serviceCreator;
			}

			public bool IsInstanceCreated
			{
				get { return _svcInstance != null; }
			}

			public object GetInstance()
			{
				if (_svcInstance == null)
					lock (this)
						if (_svcInstance == null)
						{
							_svcInstance = _serviceCreator();
							if (_svcInstance == null)
								throw new ArgumentNullException("serviceCreator", "Could not "
									+ "create instance of a service");
						}
				return _svcInstance;
			}
		}
		#endregion
	}
}
